"""# 8 node directed graph example from Robert Tarjan.

Reference:
    [Figure 3] Tarjan, R. (1972). Depth-First Search and Linear Graph Algorithms.
    SIAM Journal on Computing, 1(2), 146-160.
    [DOI: 10.1136/0201010](https://doi.org/10.1137/0201010)
"""

edge_weights = ["adjacency"]
